#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/types.h>
#include <linux/mm.h> 
#include "twinhan_linux.h"

#ifndef MAP_NR
#define  MAP_NR(addr)   (__pa(addr) >> PAGE_SHIFT)
#endif
#include "dmaBuffer.h"
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifdef DMABUFFER_DEBUG
#define PRINTK(fmt, args...) printk(fmt, ## args)
#else
#define PRINTK(fmt, args...)
#endif

#define PPRINTK(fmt, args...)
#define DMABUFFER_ORDER		1
#define DMABUFFER_BLOCKSIZE	(PAGE_SIZE * (1UL<<DMABUFFER_ORDER))
#define DMABUFFER_BLOCKS	((1UL<<(32UL-DMABUFFER_ORDER)) / PAGE_SIZE)

unsigned char *mmbuf;  //used by IP data demux
struct dmaBufferStruct
{
    caddr_t address;
    unsigned long size;
    int flags;
    struct dmaBufferStruct *next;
};
static struct dmaBufferStruct usedTail = {NULL, 0, 0, NULL};
static struct dmaBufferStruct usedHead = {NULL, 0, 0, &usedTail};
static struct dmaBufferStruct freeTail = {NULL, 0, 0, NULL};
static struct dmaBufferStruct freeHead = {NULL, 0, 0, &freeTail};
extern int brooktree_cnt;

static inline int calcNumPages(unsigned long size, unsigned long blockSize)
{
    DBGMSG(">>calcNumPages()\n");
        
    int numPages;
    numPages=size/blockSize;
    if ((size & (blockSize-1)) != 0)
    	numPages++;
    
    DBGMSG("<<calcNumPages()\n");
    return numPages;
}

static int findRegion(void * pageList[], int pageListSize, unsigned long size)
{
    DBGMSG(">>findRegion()\n");
        
    int i,j;
    int numPages;
    int startIndex;
    numPages = calcNumPages(size, DMABUFFER_BLOCKSIZE);
    startIndex=-1;
    j = 0;
    for(i = pageListSize - 1; i >= 0; i--) {
	    if(pageList[i])    
	        j++;
	    else    
	        j = 0;
	    if(j >= numPages) {
	        startIndex = i;
	        break;
	    }
    }
    
    DBGMSG("<<findRegion()\n");
    return(startIndex);
}

void mapMemory(void * pageList[], int pageListSize, int startIndex, unsigned long size)
{ 
    DBGMSG(">>mapMemory()\n");
        
    int i;
    int numPages;
    void *start_addr, *end_addr;

    numPages = calcNumPages(size, DMABUFFER_BLOCKSIZE);
    if ((startIndex >= 0) && (startIndex <= pageListSize-numPages)) {
	    start_addr=(void *)pageList[startIndex];
	    end_addr=start_addr+size-1;
	    for (i = MAP_NR(start_addr); i <= MAP_NR(end_addr); i++)
	        set_bit(PG_reserved, &mem_map[i].flags);
	    for (i=startIndex;i<startIndex+numPages;i++)
            pageList[i] = 0;
    }
    
    DBGMSG("<<mapMemory()\n");
}

static void * allocDmaBufferPages(unsigned long size) 
{
    DBGMSG(">>allocDmaBufferPages()\n");
        
    unsigned long page;
    unsigned long bufPage;
    int bufPageIndex;
    void ** pageList;
    int i;
    int nomem = 0;
    int numPages;
    int startIndex = -1;
    void *retPtr=NULL;
//MYMSG(">>allocDmaBufferPages(): DMABUFFER_BLOCKS=%ld, DMABUFFER_BLOCKSIZE=%ld, PAGE_SIZE=%ld\n", DMABUFFER_BLOCKS, DMABUFFER_BLOCKSIZE, PAGE_SIZE);
    if(NULL == (pageList = vmalloc(DMABUFFER_BLOCKS * sizeof(void *))) )     
    {
        ERRMSG("allocDmaBufferPages: pageList kmalloc fail!! \n");
        return(NULL);
    }    
    for(i=0;i<DMABUFFER_BLOCKS;i++)	
        pageList[i] = 0;
    numPages=calcNumPages(size, DMABUFFER_BLOCKSIZE);
    while(!nomem) {
	    for(i = 0; i < numPages; i++) {
#ifdef DMABUFFER_ISA
	        if(0 == (page = __get_dma_pages(GFP_KERNEL, DMABUFFER_ORDER)) )
#else            
	        if(0 == (page = __get_free_pages(GFP_KERNEL, DMABUFFER_ORDER)) )
#endif
	        {
	            ERRMSG("allocDmaBufferPages: __get_dma_pages fail!! i=%d, DMABUFFER_ORDER=%d\n", i, DMABUFFER_ORDER);
		        nomem = 1;
                break;
	        }
    	    bufPage = virt_to_bus((void *) page);
    	    bufPageIndex = bufPage / DMABUFFER_BLOCKSIZE;
    	    pageList[bufPageIndex] = (void *) page;
	    }
	    if(-1 != (startIndex = findRegion(pageList, DMABUFFER_BLOCKS, size)) )
	        break;
    }
    if(-1 != startIndex) {
	    retPtr = pageList[startIndex];
	    mapMemory(pageList, DMABUFFER_BLOCKS, startIndex, size);
    }
    for(i = 0; i < DMABUFFER_BLOCKS; i++)
        if(pageList[i] != 0)    
            free_pages((unsigned long)pageList[i], DMABUFFER_ORDER);
    vfree(pageList);
    
    DBGMSG("<<allocDmaBufferPages()\n");
    return retPtr;
}

static int freeDmaBufferPages(void * dmaBuffer, unsigned long size)
{
    DBGMSG(">>freeDmaBufferPagess()\n");
        
    int i;
    int numPages;
    if (dmaBuffer != NULL) {
    	for (i = MAP_NR(dmaBuffer); i <= MAP_NR(dmaBuffer+size-1); i++)
    	    clear_bit(PG_reserved, &mem_map[i].flags);
    	numPages=calcNumPages(size, DMABUFFER_BLOCKSIZE);
    	for(i=0;i<numPages;i++)
    	    free_pages((unsigned long)dmaBuffer+i*DMABUFFER_BLOCKSIZE, DMABUFFER_ORDER);
    }
    
    DBGMSG("<<freeDmaBufferPagess()\n");
    return 0;
}

caddr_t allocDmaBuffer(unsigned long size)
{
    DBGMSG(">>allocDmaBuffer()\n");
        
    struct dmaBufferStruct *ptr;
    void * retPtr=NULL;
    ptr= freeHead.next;
    
    while ((ptr != NULL) && (ptr != &usedTail)) {
    	if ((ptr->address != NULL) && (ptr->flags == 0) && (ptr->size >= size)) {
    	    ptr->flags++;
    	    retPtr = ptr->address;
    	    break;
    	}
    	ptr = ptr->next;
    }
    
    DBGMSG("<<allocDmaBuffer()\n");
    return retPtr;
}

caddr_t bigphysarea_alloc(unsigned long size)
{
    DBGMSG(">>bigphysarea_alloc()\n");
        
    struct dmaBufferStruct *ptr, *prev;
    struct dmaBufferStruct *c_ptr = NULL;
    struct dmaBufferStruct *new_ptr;
    prev = &freeHead;
    ptr = prev->next;
    while ((ptr != NULL)&&(ptr != &freeTail)) {
    	if (ptr->size >= size) {
            c_ptr = ptr;
    	    break;
    	}
    	prev = ptr;
    	ptr = ptr->next;
    }
    if (c_ptr == NULL)	
        return NULL;
    prev->next = c_ptr->next;
    c_ptr->next = usedHead.next;
    usedHead.next = c_ptr;
    if (c_ptr->size > size + 8192) {
    	if ((new_ptr = (struct dmaBufferStruct *)kmalloc(sizeof(struct dmaBufferStruct), GFP_KERNEL))==NULL) 
    	    return NULL;
    	new_ptr->next = prev->next;
    	prev->next = new_ptr;
    	new_ptr->address = c_ptr->address + size;
    	new_ptr->size = c_ptr->size - size;
    	c_ptr->size = size;
    }
    
    DBGMSG("<<bigphysarea_alloc()\n");
    return c_ptr->address;
}

int bigphysarea_free(caddr_t base, unsigned long size)
{
    DBGMSG(">>bigphysarea_free()\n");
        
    struct dmaBufferStruct *prev, *next, *ptr;
    struct dmaBufferStruct *c_ptr = NULL;
    prev = &usedHead;
    ptr = prev->next;
    while ((ptr != NULL)&&(ptr != &usedTail))  {
    	if (ptr->address == base){
    	    c_ptr = ptr;
    	    break;
    	}
    	prev = ptr;
    	ptr = ptr->next;
    }
    if (c_ptr == NULL)	
        return 1;
    prev->next = c_ptr->next;
    prev = &freeHead;
    ptr = prev->next;
    while ((ptr != NULL) && ( ptr != &freeTail)) {
	    if (ptr->address >= base)   break;
	    prev = ptr;
	    ptr = ptr->next;
    }
    c_ptr->next = prev->next;
    prev->next = c_ptr;
    if ((c_ptr != NULL)&&(c_ptr->next != &freeTail)&&(c_ptr->address + c_ptr->size == c_ptr->next->address)){
	    next = c_ptr->next;
	    c_ptr->size = c_ptr->size + next->size;
	    c_ptr->next = next->next;
	    kfree(next);
    }
    if ((prev != NULL)&&(prev != &freeHead)&&(prev->address + prev->size == c_ptr->address)) {
	    prev->size = prev->size + c_ptr->size;
	    prev->next = c_ptr->next;
	    kfree(c_ptr);
    }
    
    DBGMSG("<<bigphysarea_free()\n");
    return 0;
}

static unsigned long int size=3060; //dmaBuffer size for 1 card
int init_dmaBuffer(void)
{
    DBGMSG(">>init_dmaBuffer()\n");    
    
    void *ptr;
    struct dmaBufferStruct *sptr;
    int numPages;
    size=3060*brooktree_cnt;
    if (size == 0)  {
        TRYMSG("init_dmaBuffer()=> brooktree_cnt=%d, size=%ld \n", brooktree_cnt, size);    
        return -EINVAL;
    }    
    numPages = calcNumPages(size*1024, DMABUFFER_BLOCKSIZE);
    if ((ptr=allocDmaBufferPages(size*1024))==NULL)    	
        return -EIO;
    if((sptr =	(struct dmaBufferStruct *) kmalloc(sizeof(struct dmaBufferStruct), GFP_KERNEL))==NULL) {
	    freeDmaBufferPages(ptr, size*1024);
	    return -EIO;
    }
    mmbuf=ptr;  //IP
    sptr->next = freeHead.next;
    freeHead.next = sptr;
    sptr->address = ptr;
    sptr->size = numPages*DMABUFFER_BLOCKSIZE;
    
    DBGMSG("<<init_dmaBuffer()\n");
    return 0;
}

int bigphysarea_info()
{
  DBGMSG(">>bigphysarea_info()\n");
        
  struct dmaBufferStruct *ptr;
  int count;
  count = 0;
  ptr = usedHead.next;
  while((count < 500)&&(ptr != NULL)&&(ptr != &usedTail)) {
    ptr = ptr->next;
    count++;
  }
  count = 0;
  ptr = freeHead.next;
  while((count < 500)&&(ptr != NULL)&&(ptr != &freeTail))  {
    ptr = ptr->next;
    count++;
  }
  
  DBGMSG("<<bigphysarea_info()\n");
  return 0;
}

void release_dmaBuffer(void)
{
    DBGMSG(">>release_dmaBuffer()\n");
        
    struct dmaBufferStruct *ptr;
    struct dmaBufferStruct *ptr2;
    ptr=usedHead.next;
    while ((ptr != NULL)&&(ptr != &usedTail)) {
    	if (ptr->address != NULL)    
    	    freeDmaBufferPages(ptr->address,ptr->size);
    	ptr2 = ptr;
    	ptr = ptr->next;
    	kfree(ptr2);
    }
    ptr=freeHead.next;
    while ((ptr != NULL)&&(ptr != &freeTail)) {
    	if (ptr->address != NULL)  
    	    freeDmaBufferPages(ptr->address,ptr->size);
    	ptr2 = ptr;
    	ptr = ptr->next;
    	kfree(ptr2);
    }
    
    DBGMSG("<<release_dmaBuffer()\n");
}

